import java.util.*;

/**
 * File information:<br>
 * $Author: nmk148 $<br>
 * First check-in: $Date: 2003/04/11 17:49:15 $<br>
 * $Revision: 1.4 $<br>
 * <br>
 * This class represents the whole population and manages the different
 * steps of a generation transition.
 **/
public class Population {


	
	private Random random; /* Random number generator for this population */
	private Problem problem; /* The problem of this population */
	
	private Individual[] pop; /* The individuals in a generation */
	private Individual best; /* The best invidividual (for the elitist scheme) */
	
	int sizeGA; /* The number of GA Individuals in this population */
	int sizeASGA; /* The number of ASGA Individuals in this population */
	private int birth; /* The number of children per individual */
	private double mutate;/* The number of individuals to 
				 mutate per generation */
	private double matingRate; /* The probability for an individual to
				      mate */

	private LinkedList childrenGA;    /* for the children */
	private LinkedList childrenASGA;  /* dito */
	private LinkedList mater; /* A linked list for mater */

	private double avgFitness; /* The average fitness */
	private double stdDevFitness;/* The standard deviation of the 
					fitness distribution */

	
	/**
	 * Creates a new population with randomly initialized
	 * individuals to solve a problem
	 * @param sizeGA the number of GA individuals
	 * @param sizeASGA the number of ASGA individuals
	 * @param survival survival rate for ASGA individuals at different
	 *              differen ages
	 * @param mate the probalitiy that an individual performs a crossover.
	 * @param birth the number of children per individual
	 * @param mutate the probability that an individual mutates.
	 * @param p the problem
	 **/
	public Population(int sizeGA,
			  int sizeASGA,
			  double[] survival,
			  double mate, 
			  int birth,
			  double mutate,
			  Problem p) {
		
		random = new Random();
		this.sizeGA = sizeGA;
		this.sizeASGA = sizeASGA;
		this.problem = p;
		this.birth = birth;
		this.mutate = mutate;
		this.matingRate = mate;
		pop = new Individual[sizeGA+sizeASGA];
		for (int i=0; i<sizeGA; i++) { /* Create GA Invidiual */
			pop[i] = new Individual(p, birth,
					survival, false, true);
		}
		
		for (int i=0; i<sizeASGA; i++) { /* Create ASGA Individual */
			pop[i+sizeGA] = new Individual(p, birth, 
					survival, true, true);
		}
	}

	

	/** 
	 * Performs the transition from one generation to another.
	 * This transition consists of:<br>
	 * <ol>
	 *   <li>Crossover: see crossover()</li>
	 *   <li>Mutation: see mutate()</li>
	 *   <li>Evaluation: see evaluate()</li>
	 *   <li>Selection: see select()</li>
	 * </ol>
	 **/
	public void nextGeneration() {

		/* We seperate ASGA and GA Children */
		childrenASGA = new LinkedList();
		childrenGA = new LinkedList();
		mater = new LinkedList();
		
		if (Debug.level > 0) Debug.log("Dividing");
		divide();
		if (Debug.level > 0) Debug.log("Crossing");
		crossover();
		if (Debug.level > 0) Debug.log("Mutating");
		mutate();
		if (Debug.level > 0) Debug.log("Evaluating");
		evaluate();
		if (Debug.level > 0) Debug.log("Selecting");
		select();
		if (Debug.level > 0) Debug.log("done");
	}

	

	/** 
	 * Divides the population into two groups:
	 * Maters and non-mater. Everything that doesn't mate is
	 * put into the children list for the next population.
	 * Everything else performs a crossover.
	 **/
	protected void divide() {
	
		/* Save best individual */
		best = new Individual(getBest());

		int age = 0;  /* count for ASGA */
		int noAge = 0 ; /* count for GA */
		/* Divide individuals */
		for (int i=0; i<pop.length; i++) {
			boolean mate; /* does this individual mate? */
			
			/* Let this individual get older */
			pop[i].age(); 

			if (pop[i].asga()) {
		//		mate = random.nextDouble() < matingRate;
				mate = age++ > (1-matingRate)*sizeASGA;
					
			} else {
		                mate = noAge++ > (1-matingRate)*sizeGA;
			// 	mate = random.nextDouble() < matingRate;
			}
				
			/* Does this individual mate? */
			if (mate) {
				mater.add(pop[i]);
			}

			/* Does this individual survive? */
			if (pop[i].survive(mate)) {
				if (pop[i].asga()) {
					childrenASGA.add(pop[i]);
				} else {
					childrenGA.add(pop[i]);
				}
			}
			
		}
		/* If there are less than 2 mater, we have a problem.
		 * There can not be a crossover...
		 * This should not happen, or at least very rarely.
		 * But it could happen and then we've got a problem
		 * in selection (because there may be to few individuals
		 * and we get a ArrayOutOfBoundsException...
		 * For now, we ensure that there are enought individuals 
		 * by putting this one back into its children list.
		 */
		if (mater.size() == 1) {
			Individual ind = (Individual)mater.removeLast();
			if (ind.asga()) {
				childrenASGA.add(ind);
			} else {
				childrenGA.add(ind);
			}
		}
	}



	/**
	 * Performs a crossover for a certain number of individuals.
	 * The number is the <i>mate</i> parameter of the constructor.
	 * Performs the following actions to do that:
	 * <ol>
	 *   <li>Copy each individual with a probability of <i>mate</i>
	 *       into a datastructure. Put all others into the <i>children</i>
	 *       list.</li>
	 *   <li>Take the first one and perform <i>child</i> crossovers
	 *       with random individuals and put children into the 
	 *       <i>children</i> list</li>
	 *   <li>Remove all individuals that cannot bear children anymore</li>
	 *   <li>Perform steps 2 and 3 until one or zero individuals 
	 *       are left</li>
	 * </ol>
	 **/ 
	private void crossover() {
		/* Debug */ if (Debug.on) Debug.log("M: "+mater); 
		/* Debug */ if (Debug.on) Debug.log("GA: "+childrenGA); 
		/* Debug */ if (Debug.on) Debug.log("ASGA: "+childrenASGA);
		if (Debug.level > 0) 
			Debug.assert(mater.size() != 1, "Invalid maters size");
		if (Debug.level > 0) 
			Debug.assert(mater.size()+childrenGA.size()>=sizeGA,
					"Invalid maters size"+mater.size()
					+","+childrenGA.size());
		while (mater.size() > 1) {
			Individual parent1, parent2, child1, child2;
			
			int partner = random.nextInt(mater.size()-1) + 1;
			parent1 = (Individual)mater.getFirst();
			parent2 = (Individual)mater.get(partner);
			
			/* gs: Which gene should be switched? */
			int gs = random.nextInt(problem.geneSize());
			boolean switchASGA = (random.nextDouble() < 0.1);
			child1 = parent1.crossover(gs, switchASGA, parent2);
			child2 = parent2.crossover(gs, switchASGA, parent1);
		
			if (Debug.level > 0) 
				Debug.assert(parent1.asga() == child1.asga() 
						|| parent2.asga() == child1.asga(),
						"Invalid children");
			if (Debug.level > 0) 
				Debug.assert(parent1.asga() == child2.asga() 
						|| parent2.asga() == child2.asga(),
						"Invalid children");
			if (child1.asga()) {
				childrenASGA.add(child1);
			} else {
				childrenGA.add(child1);
			}
			if (child2.asga()) {
				childrenASGA.add(child2);
			} else {
				childrenGA.add(child2);
			}

			/* remove exhausted parents */
			/* Debug */ if (Debug.level > 3) Debug.log(""+partner); 
			if (parent2.outOfEnergy()) mater.remove(partner);
			if (parent1.outOfEnergy()) mater.removeFirst();
		}
		if (Debug.level > 0) 
			Debug.assert(sizeGA <= childrenGA.size(),
					"Too few childrenGA!"+sizeGA+">"
					+childrenGA.size());

		if (Debug.level > 0) 
			Debug.assert(sizeASGA <= childrenASGA.size(),
					"Too few childrenASGA!"+sizeASGA+">"
					+childrenASGA.size());
	}



	/**
	 * Mutates each individual in the <i>children</i> list
	 * with a probability of 
	 * <i>mutate</i>.
	 **/
	private void mutate() {
		Iterator iter = childrenGA.iterator();
		while(iter.hasNext()) {
			Individual ind = (Individual)iter.next();
			if (random.nextDouble() < mutate) {
				ind.mutate();
			}
		}
		
		iter = childrenASGA.iterator();
		while(iter.hasNext()) {
			Individual ind = (Individual)iter.next();
			if (random.nextDouble() < mutate) {
				ind.mutate();
			}
		}
	}


	/**
	 * Sorts the children by their fitness 
	 **/
	private void evaluate() {
		/* Insert best invidual */
		if (best.asga()) {
			childrenASGA.add(best);
		} else {
			childrenGA.add(best);
		}
		Collections.sort(childrenGA);
		Collections.sort(childrenASGA);
	}

	/**
	 * Moves the fittest individuals from the sorted children 
	 * list into the population array.
	 **/
	private void select() {
		if (Debug.level > 0) 
			Debug.assert(sizeGA <= childrenGA.size(),
					"Too few childrenGA! "+sizeGA+">"
					+childrenGA.size());

		for (int i=0; i<sizeGA; i++) {
			pop[i] = (Individual)childrenGA.removeLast();
		}
		if (Debug.level > 0) 
			Debug.assert(sizeASGA <= childrenASGA.size(),
					"Too few childrenASGA! "+sizeASGA+">"
					+childrenASGA.size());
		for (int i=0; i<sizeASGA; i++) {
			pop[sizeGA+i] = (Individual)childrenASGA.removeLast();
		}
		
	}



	/** 
	 * Causes the statistics to be computed 
	 */
	public void computeStats() {

		/* Average fitness */
		double sum = 0;
		for (int i=0; i<pop.length; i++) {
			sum += pop[i].fitness();
		}
		avgFitness = sum / pop.length;

		/* Standard deviation */
		sum=0;
		for (int i=0; i<pop.length; i++) {
			sum += (pop[i].fitness() - avgFitness)
				* (pop[i].fitness() - avgFitness);
		}
		stdDevFitness = Math.sqrt(sum / pop.length);
	}


	/**
	 * Returns the individual with the best fitness
	 * @return best individual
	 **/
	public Individual getBest() {
		return pop[0];
	}
	
	/**
	 * Returns the average fitness
	 * @return the average fitness
	 **/
	public double getAverageFitness() {
		return avgFitness;
	}

	/**
	 * Returns the standard deviation of the fitness distribution
	 * @return the standard deviation
	 **/
	public double getStdDevFitness() {
		return stdDevFitness;
	}


	/**
	 * Prints the whole population to stderr
	 **/
	public void printAll() {
		for (int i=0; i<sizeGA+sizeASGA; i++) {
			System.err.println(pop[i]);
		} 

	}


}



/*
 * History of this file:
 * $Log: Population.java,v $
 * Revision 1.4  2003/04/11 17:49:15  nmk148
 * Done, handed in everything...
 *
 * Revision 1.3  2003/04/09 16:44:48  nmk148
 * Done testing
 *
 * Revision 1.2  2003/03/28 16:12:22  nmk148
 * Finished
 *
 * Revision 1.1  2003/03/27 02:09:49  nmk148
 * Finished
 *
 * Revision 1.2  2003/03/23 17:19:13  nmk148
 * Added statistical functions and stuff
 *
 * Revision 1.1  2003/03/17 07:35:04  nmk148
 * First checkin
 *
 */
